@@ -3,6 +3,7 @@
 import os
 
 from enum import Enum
+from math import ceil
 from struct import pack
 from ubuntu_image.helpers import run
 
@@ -167,6 +168,13 @@ class MBRImage(Image):
         for key, value in sfdisk_args.items():
             if key == 'new':
                 offset, size = value.split(':')
+                # XXX: special behavior from sfdisk: it rounds our partition
+                # size down to the nearest MiB without asking.  So round up,
+                # and if it extends past the end of the disk sfdisk will round
+                # it down to where it was supposed to be in the first place.
+                if size.endswith('K'):
+                    size = int(size.rstrip('K').lstrip('+'))
+                    size = '+{}M'.format(ceil(size / 1024))
                 command_input.extend([
                     'start={}'.format(offset),
                     'size={}'.format(size),
@@ -177,6 +185,14 @@ class MBRImage(Image):
                 if isinstance(value, tuple):
                     value = value[0]
                 command_input.append('type={}'.format(value))
+            elif key == 'change_name':
+                # If the gadget.yaml has an mbr schema with a named structure,
+                # the builder will try to set the partition's name to this
+                # value.  That's not supported by sfdisk and is probably not
+                # allowed by the mbr partition table format.  Rather than
+                # change the builder (i.e. in case we find a workaround), for
+                # now, just ignore the argument.
+                pass
             else:
                 raise ValueError('{} option not supported for MBR partitions'
                                  .format(key))
